"""Main event stream pane for the read-only TUI prototype."""

from hashlib import sha1

from msgspec import Struct
from rich.console import Group, RenderableType
from rich.markdown import Markdown
from rich.panel import Panel
from rich.style import Style
from rich.table import Table
from rich.text import Text
from textual.events import Click, Key, Resize
from textual.message import Message
from textual.widgets import RichLog

from aceai.agent.tui.events import TUIEvent, TUIEventKind, TUIIdeaItem
from aceai.agent.tui.state import TUIRunState
from aceai.agent.tui.theme import EVENT_LABELS, EVENT_STYLES

PROMPT_BAR_STYLE = "bold #eceff4 on #3b4252"
PROMPT_MARK_STYLE = "bold #88c0d0 on #3b4252"
SUBTLE_BULLET_STYLE = "bold #9aa3b2"
REASONING_MARK_STYLE = "bold #d08770"
EXPAND_MARK_STYLE = "bold #88c0d0"
TRANSCRIPT_GUTTER = "  "


class StreamWidget(RichLog):
    """Render the readable event transcript."""

    class EventSelected(Message):
        def __init__(self, event_id: str) -> None:
            self.event_id = event_id
            super().__init__()

    DEFAULT_CSS = """
    StreamWidget {
        border: solid #81a1c1;
        background: #2e3440;
        color: #e5e9f0;
        padding: 0 1;
        width: 1fr;
        height: 1fr;
        overflow-y: auto;
        overflow-x: hidden;
    }
    """

    def __init__(
        self,
        state: TUIRunState | None = None,
        *,
        id: str | None = None,
    ) -> None:
        super().__init__(id=id, wrap=True, auto_scroll=True, min_width=0)
        self._state = state or TUIRunState()
        self._debug_mode = False
        self._selected_debug_index = 0
        self._debug_line_spans: list[_DebugLineSpan] = []
        self._expanded_tool_activity_ids: set[str] = set()
        self._tool_activity_line_spans: list[_ToolActivityLineSpan] = []

    @property
    def debug_mode(self) -> bool:
        return self._debug_mode

    def set_debug_mode(self, enabled: bool) -> str | None:
        self._debug_mode = enabled
        selectable_events = _selectable_debug_events(self._state.events)
        if enabled and selectable_events:
            self._selected_debug_index = min(
                self._selected_debug_index,
                len(selectable_events) - 1,
            )
            selected_event_id = selectable_events[self._selected_debug_index].event_id
        else:
            selected_event_id = None
        self.set_state(self._state)
        return selected_event_id

    def set_state(self, state: TUIRunState) -> None:
        self._state = state
        self.clear()
        if not self._state.events:
            self.write(Text("No events yet", style="#d8dee9"))
        elif self._debug_mode:
            self._write_debug_stream()
        else:
            self._tool_activity_line_spans = []
            for entry in _render_event_entries(
                self._state.events,
                collapse_tool_activity=self._state.status == "completed",
                expanded_tool_activity_ids=self._expanded_tool_activity_ids,
            ):
                start_line = len(self.lines)
                self._write_stream_renderable(entry.renderable)
                if entry.tool_activity_id != "":
                    self._tool_activity_line_spans.append(
                        _ToolActivityLineSpan(
                            activity_id=entry.tool_activity_id,
                            start_line=start_line,
                            end_line=len(self.lines),
                        )
                    )
        self.call_after_refresh(self.scroll_end, animate=False)

    def on_key(self, event: Key) -> None:
        if not self._debug_mode:
            return
        if event.key in ("up", "k"):
            self._select_debug_event(self._selected_debug_index - 1)
            event.stop()
            return
        if event.key in ("down", "j"):
            self._select_debug_event(self._selected_debug_index + 1)
            event.stop()
            return
        if event.key == "home":
            self._select_debug_event(0)
            event.stop()
            return
        if event.key == "end":
            self._select_debug_event(len(_selectable_debug_events(self._state.events)) - 1)
            event.stop()

    def on_click(self, event: Click) -> None:
        line_index = self.scroll_y + event.y
        if not self._debug_mode:
            activity_id = _tool_activity_id_from_click(event)
            if activity_id != "":
                self._toggle_tool_activity(activity_id)
                event.stop()
                return
            for span in self._tool_activity_line_spans:
                if span.start_line <= line_index < span.end_line:
                    self._toggle_tool_activity(span.activity_id)
                    event.stop()
                    return
            return
        for span in self._debug_line_spans:
            if span.start_line <= line_index < span.end_line:
                self._select_debug_event(span.index)
                event.stop()
                return

    def _toggle_tool_activity(self, activity_id: str) -> None:
        if activity_id in self._expanded_tool_activity_ids:
            self._expanded_tool_activity_ids.remove(activity_id)
        else:
            self._expanded_tool_activity_ids.add(activity_id)
        self.set_state(self._state)

    def _select_debug_event(self, index: int) -> None:
        selectable_events = _selectable_debug_events(self._state.events)
        if not selectable_events:
            return
        self._selected_debug_index = max(0, min(index, len(selectable_events) - 1))
        self.set_state(self._state)
        self.post_message(
            self.EventSelected(selectable_events[self._selected_debug_index].event_id)
        )

    def _write_debug_stream(self) -> None:
        self._debug_line_spans = []
        selectable_events = _selectable_debug_events(self._state.events)
        if not selectable_events:
            self.write(Text("No inspectable messages yet", style="#d8dee9"))
            return
        self._selected_debug_index = min(
            self._selected_debug_index,
            len(selectable_events) - 1,
        )
        selected_event_id = selectable_events[self._selected_debug_index].event_id
        for index, entry in enumerate(_render_debug_events(selectable_events)):
            renderable = entry.renderable
            if entry.event_id == selected_event_id:
                renderable = _selected_debug_panel(renderable)
            elif index > 0:
                self.write(Text(""))
            start_line = len(self.lines)
            self._write_stream_renderable(renderable)
            self._debug_line_spans.append(
                _DebugLineSpan(
                    index=index,
                    event_id=entry.event_id,
                    start_line=start_line,
                    end_line=len(self.lines),
                )
            )

    def on_resize(self, event: Resize) -> None:
        if self._state.events:
            self.set_state(self._state)

    def _write_stream_renderable(self, renderable: RenderableType) -> None:
        if isinstance(renderable, Table):
            self.write(renderable, expand=True)
            return
        self.write(renderable, width=_available_stream_width(self))


class _ToolBlockState(Struct, kw_only=True):
    call_id: str
    name: str | None = None
    arguments: str = ""
    output: str = ""
    status: str = "running"


class _WorkingHistoryItem(Struct, kw_only=True):
    renderable: RenderableType | None = None
    tool_block: _ToolBlockState | None = None


class _ToolActivityState(Struct, kw_only=True):
    items: list[_WorkingHistoryItem]


class _StreamRenderable(Struct, kw_only=True):
    renderable: RenderableType
    tool_activity_id: str = ""


class _DebugRenderable(Struct, kw_only=True):
    event_id: str
    renderable: RenderableType


class _ToolActivityLineSpan(Struct, kw_only=True):
    activity_id: str
    start_line: int
    end_line: int


class _DebugLineSpan(Struct, kw_only=True):
    index: int
    event_id: str
    start_line: int
    end_line: int


def _render_events(
    events: list[TUIEvent],
    *,
    collapse_tool_activity: bool = False,
) -> list[RenderableType]:
    return [
        entry.renderable
        for entry in _render_event_entries(
            events,
            collapse_tool_activity=collapse_tool_activity,
            expanded_tool_activity_ids=set(),
        )
    ]


def _render_event_entries(
    events: list[TUIEvent],
    *,
    collapse_tool_activity: bool,
    expanded_tool_activity_ids: set[str],
) -> list[_StreamRenderable]:
    if collapse_tool_activity:
        return _render_completed_event_entries(
            events,
            expanded_tool_activity_ids,
        )

    entries: list[_StreamRenderable] = []
    assistant_buffer = ""
    assistant_buffer_step_id = ""
    thinking_buffer = ""
    assistant_step_ids: set[str] = set()
    pending_reasoning: dict[str, list[TUIEvent]] = {}
    rendered_reasoning_step_ids: set[str] = set()
    tool_blocks: dict[str, _ToolBlockState] = {}
    rendered_tool_call_ids: set[str] = set()
    pending_working_history: _ToolActivityState | None = None

    for event in events:
        if event.kind == "reasoning_summary":
            pending_reasoning.setdefault(event.step_id, []).append(event)
            continue
        if event.kind == "assistant_delta":
            assistant_buffer_step_id = event.step_id
            assistant_buffer += event.content
            continue
        if event.kind == "thinking_delta":
            thinking_buffer += event.content
            continue

        if event.tool_call_id is not None and event.kind in (
            "tool_call_delta",
            "tool_started",
            "tool_output",
            "tool_approval_requested",
            "tool_approval_resolved",
            "tool_completed",
            "tool_failed",
            "run_suspended",
        ):
            thinking_buffer = _flush_thinking_buffer(entries, thinking_buffer)
            _flush_pending_reasoning(
                entries,
                event.step_id,
                pending_reasoning,
                rendered_reasoning_step_ids,
            )
            assistant_buffer, assistant_buffer_step_id = _flush_assistant_buffer(
                entries,
                assistant_buffer,
                assistant_buffer_step_id,
                assistant_step_ids,
            )
            _update_tool_block(tool_blocks, event)
            if event.kind in ("tool_completed", "tool_failed", "tool_approval_requested"):
                entries.append(
                    _StreamRenderable(
                        renderable=_render_tool_block(tool_blocks[event.tool_call_id])
                    )
                )
                rendered_tool_call_ids.add(event.tool_call_id)
            continue

        if event.kind == "llm_completed":
            _flush_pending_reasoning(
                entries,
                event.step_id,
                pending_reasoning,
                rendered_reasoning_step_ids,
            )
            if event.tool_calls:
                if assistant_buffer_step_id == event.step_id:
                    assistant_buffer = ""
                    assistant_buffer_step_id = ""
                continue
        if _is_invisible_control_event(event):
            continue
        thinking_buffer = _flush_thinking_buffer(entries, thinking_buffer)
        assistant_buffer, assistant_buffer_step_id = _flush_assistant_buffer(
            entries,
            assistant_buffer,
            assistant_buffer_step_id,
            assistant_step_ids,
        )
        pending_working_history = _flush_tool_activity(
            entries,
            pending_working_history,
            expanded_tool_activity_ids,
        )

        if event.kind == "llm_completed" and event.step_id in assistant_step_ids:
            continue

        rendered = _render_event(event)
        if rendered is not None:
            _flush_pending_reasoning(
                entries,
                event.step_id,
                pending_reasoning,
                rendered_reasoning_step_ids,
            )
            if event.kind == "user_message" and entries:
                entries.append(_StreamRenderable(renderable=Text("")))
            if event.kind == "llm_completed":
                assistant_step_ids.add(event.step_id)
            entries.append(_StreamRenderable(renderable=rendered))

    thinking_buffer = _flush_thinking_buffer(entries, thinking_buffer)
    for step_id in pending_reasoning:
        _flush_pending_reasoning(
            entries,
            step_id,
            pending_reasoning,
            rendered_reasoning_step_ids,
        )
    for call_id, tool_block in tool_blocks.items():
        if call_id in rendered_tool_call_ids:
            continue
        if tool_block.name is None:
            continue
        entries.append(_StreamRenderable(renderable=_render_tool_block(tool_block)))
    _flush_tool_activity(
        entries,
        pending_working_history,
        expanded_tool_activity_ids,
    )
    assistant_buffer, assistant_buffer_step_id = _flush_assistant_buffer(
        entries,
        assistant_buffer,
        assistant_buffer_step_id,
        assistant_step_ids,
    )
    return entries


def _render_completed_event_entries(
    events: list[TUIEvent],
    expanded_tool_activity_ids: set[str],
) -> list[_StreamRenderable]:
    entries: list[_StreamRenderable] = []
    assistant_buffer = ""
    assistant_buffer_step_id = ""
    assistant_step_ids: set[str] = set()
    thinking_buffer = ""
    pending_working_history: _ToolActivityState | None = None
    tool_blocks: dict[str, _ToolBlockState] = {}
    rendered_tool_call_ids: set[str] = set()

    for event in events:
        if event.kind == "user_message":
            pending_working_history = _flush_thinking_buffer_to_working_history(
                pending_working_history,
                thinking_buffer,
            )
            thinking_buffer = ""
            pending_working_history, assistant_buffer, assistant_buffer_step_id = (
                _flush_completed_turn(
                    entries,
                    pending_working_history,
                    assistant_buffer,
                    assistant_buffer_step_id,
                    assistant_step_ids,
                    expanded_tool_activity_ids,
                )
            )
            if entries:
                entries.append(_StreamRenderable(renderable=Text("")))
            rendered = _render_event(event)
            if rendered is not None:
                entries.append(_StreamRenderable(renderable=rendered))
            continue

        if event.kind == "thinking_delta":
            thinking_buffer += event.content
            continue

        if event.kind == "reasoning_summary":
            rendered = _render_event(event)
            if rendered is not None:
                pending_working_history = _append_working_history_renderable(
                    pending_working_history,
                    rendered,
                )
            continue

        if event.kind == "assistant_delta":
            assistant_buffer_step_id = event.step_id
            assistant_buffer += event.content
            continue

        if event.kind == "llm_completed":
            if event.tool_calls:
                if assistant_buffer_step_id == event.step_id:
                    assistant_buffer = ""
                    assistant_buffer_step_id = ""
                continue
            if assistant_buffer == "" and event.content != "":
                assistant_buffer_step_id = event.step_id
                assistant_buffer = event.content
            continue

        if event.kind == "run_completed":
            if assistant_buffer == "" and event.content != "":
                assistant_buffer_step_id = event.step_id
                assistant_buffer = event.content
            continue

        if event.tool_call_id is not None and event.kind in (
            "tool_call_delta",
            "tool_started",
            "tool_output",
            "tool_approval_requested",
            "tool_approval_resolved",
            "tool_completed",
            "tool_failed",
            "run_suspended",
        ):
            pending_working_history = _flush_thinking_buffer_to_working_history(
                pending_working_history,
                thinking_buffer,
            )
            thinking_buffer = ""
            _update_tool_block(tool_blocks, event)
            if event.kind in ("tool_completed", "tool_failed", "tool_approval_requested"):
                pending_working_history = _append_tool_block(
                    entries,
                    pending_working_history,
                    tool_blocks[event.tool_call_id],
                    expanded_tool_activity_ids,
                )
                rendered_tool_call_ids.add(event.tool_call_id)
            continue

        if event.kind in ("session_notice", "idea_list"):
            pending_working_history = _flush_thinking_buffer_to_working_history(
                pending_working_history,
                thinking_buffer,
            )
            thinking_buffer = ""
            pending_working_history, assistant_buffer, assistant_buffer_step_id = (
                _flush_completed_turn(
                    entries,
                    pending_working_history,
                    assistant_buffer,
                    assistant_buffer_step_id,
                    assistant_step_ids,
                    expanded_tool_activity_ids,
                )
            )
            rendered = _render_event(event)
            if rendered is not None:
                entries.append(_StreamRenderable(renderable=rendered))

    pending_working_history = _flush_thinking_buffer_to_working_history(
        pending_working_history,
        thinking_buffer,
    )
    thinking_buffer = ""
    for call_id, tool_block in tool_blocks.items():
        if call_id in rendered_tool_call_ids:
            continue
        if tool_block.name is None:
            continue
        pending_working_history = _append_tool_block(
            entries,
            pending_working_history,
            tool_block,
            expanded_tool_activity_ids,
        )
    _flush_completed_turn(
        entries,
        pending_working_history,
        assistant_buffer,
        assistant_buffer_step_id,
        assistant_step_ids,
        expanded_tool_activity_ids,
    )
    return entries


def _flush_completed_turn(
    entries: list[_StreamRenderable],
    pending_working_history: _ToolActivityState | None,
    assistant_buffer: str,
    assistant_buffer_step_id: str,
    assistant_step_ids: set[str],
    expanded_tool_activity_ids: set[str],
) -> tuple[_ToolActivityState | None, str, str]:
    pending_working_history = _flush_tool_activity(
        entries,
        pending_working_history,
        expanded_tool_activity_ids,
    )
    assistant_buffer, assistant_buffer_step_id = _flush_assistant_buffer(
        entries,
        assistant_buffer,
        assistant_buffer_step_id,
        assistant_step_ids,
    )
    return pending_working_history, assistant_buffer, assistant_buffer_step_id


def _render_debug_events(events: list[TUIEvent]) -> list[_DebugRenderable]:
    renderables: list[_DebugRenderable] = []
    for event in events:
        rendered = _render_debug_event(event)
        if rendered is None:
            continue
        renderables.append(
            _DebugRenderable(event_id=event.event_id, renderable=rendered)
        )
    return renderables


def _render_debug_event(event: TUIEvent) -> RenderableType | None:
    if event.kind == "tool_started":
        return _render_text_block(
            event.tool_name or EVENT_LABELS[event.kind],
            "running",
            event_kind=event.kind,
        )
    return _render_event(event)


def _selectable_debug_events(events: list[TUIEvent]) -> list[TUIEvent]:
    selectable: list[TUIEvent] = []
    for event in events:
        if _render_debug_event(event) is None:
            continue
        selectable.append(event)
    return selectable


def _selected_debug_panel(renderable: RenderableType) -> RenderableType:
    return Panel(
        renderable,
        border_style="#88c0d0",
        style="on #3b4252",
        padding=(0, 0),
    )


def _flush_pending_reasoning(
    renderables: list[_StreamRenderable],
    step_id: str,
    pending_reasoning: dict[str, list[TUIEvent]],
    rendered_reasoning_step_ids: set[str],
) -> None:
    if step_id in rendered_reasoning_step_ids:
        return
    events = pending_reasoning.get(step_id)
    if events is None:
        return
    for event in events:
        rendered = _render_event(event)
        if rendered is not None:
            renderables.append(_StreamRenderable(renderable=rendered))
    rendered_reasoning_step_ids.add(step_id)


def _flush_pending_reasoning_to_working_history(
    pending_working_history: _ToolActivityState | None,
    step_id: str,
    pending_reasoning: dict[str, list[TUIEvent]],
    rendered_reasoning_step_ids: set[str],
) -> _ToolActivityState | None:
    if step_id in rendered_reasoning_step_ids:
        return pending_working_history
    events = pending_reasoning.get(step_id)
    if events is None:
        return pending_working_history
    for event in events:
        rendered = _render_event(event)
        if rendered is not None:
            pending_working_history = _append_working_history_renderable(
                pending_working_history,
                rendered,
            )
    rendered_reasoning_step_ids.add(step_id)
    return pending_working_history


def _flush_assistant_buffer(
    renderables: list[_StreamRenderable],
    assistant_buffer: str,
    assistant_buffer_step_id: str,
    assistant_step_ids: set[str],
) -> tuple[str, str]:
    if assistant_buffer == "":
        return "", assistant_buffer_step_id
    renderables.append(
        _StreamRenderable(renderable=_render_assistant_block(assistant_buffer))
    )
    assistant_step_ids.add(assistant_buffer_step_id)
    return "", ""


def _flush_thinking_buffer(
    renderables: list[_StreamRenderable],
    thinking_buffer: str,
) -> str:
    if thinking_buffer:
        renderables.append(
            _StreamRenderable(
                renderable=_render_text_block(
                    "reasoning",
                    thinking_buffer,
                    event_kind="thinking_delta",
                )
            )
        )
    return ""


def _flush_thinking_buffer_to_working_history(
    pending_working_history: _ToolActivityState | None,
    thinking_buffer: str,
) -> _ToolActivityState | None:
    if thinking_buffer == "":
        return pending_working_history
    return _append_working_history_renderable(
        pending_working_history,
        _render_text_block(
            "reasoning",
            thinking_buffer,
            event_kind="thinking_delta",
        ),
    )


def _append_working_history_renderable(
    pending_working_history: _ToolActivityState | None,
    renderable: RenderableType,
) -> _ToolActivityState:
    if pending_working_history is None:
        pending_working_history = _ToolActivityState(items=[])
    pending_working_history.items.append(
        _WorkingHistoryItem(renderable=renderable)
    )
    return pending_working_history


def _append_tool_block(
    renderables: list[_StreamRenderable],
    pending_tool_activity: _ToolActivityState | None,
    tool_block: _ToolBlockState,
    expanded_tool_activity_ids: set[str],
) -> _ToolActivityState | None:
    if not _tool_block_belongs_to_activity(tool_block):
        pending_tool_activity = _flush_tool_activity(
            renderables,
            pending_tool_activity,
            expanded_tool_activity_ids,
        )
        renderables.append(_StreamRenderable(renderable=_render_tool_block(tool_block)))
        return pending_tool_activity
    if pending_tool_activity is None:
        pending_tool_activity = _ToolActivityState(items=[])
    cloned = _clone_tool_block(tool_block)
    for index, item in enumerate(pending_tool_activity.items):
        if item.tool_block is not None and item.tool_block.call_id == cloned.call_id:
            pending_tool_activity.items[index] = _WorkingHistoryItem(tool_block=cloned)
            return pending_tool_activity
    pending_tool_activity.items.append(_WorkingHistoryItem(tool_block=cloned))
    return pending_tool_activity


def _flush_tool_activity(
    renderables: list[_StreamRenderable],
    pending_tool_activity: _ToolActivityState | None,
    expanded_tool_activity_ids: set[str],
) -> None:
    if (
        pending_tool_activity is not None
        and _working_history_tool_blocks(pending_tool_activity)
    ):
        renderables.extend(
            _render_tool_activity_entries(
                pending_tool_activity,
                expanded_tool_activity_ids,
            )
        )
    return None


def _tool_block_belongs_to_activity(tool_block: _ToolBlockState) -> bool:
    if tool_block.name is None:
        raise ValueError("tool block must include a tool name before rendering")
    return tool_block.status in ("awaiting_approval", "completed", "failed")


def _clone_tool_block(tool_block: _ToolBlockState) -> _ToolBlockState:
    return _ToolBlockState(
        call_id=tool_block.call_id,
        name=tool_block.name,
        status=tool_block.status,
        arguments=tool_block.arguments,
        output=tool_block.output,
    )


def _update_tool_block(
    tool_blocks: dict[str, _ToolBlockState],
    event: TUIEvent,
) -> None:
    if event.tool_call_id is None:
        raise ValueError("tool event must include tool_call_id")
    tool_block = tool_blocks.get(event.tool_call_id)
    if tool_block is None:
        tool_block = _ToolBlockState(call_id=event.tool_call_id)
        tool_blocks[event.tool_call_id] = tool_block
    if event.tool_name is not None:
        tool_block.name = event.tool_name
    if event.tool_call is not None and event.tool_call.arguments != "":
        tool_block.arguments = event.tool_call.arguments
    elif event.kind == "tool_call_delta":
        tool_block.arguments += event.content
    if event.kind == "tool_output":
        tool_block.output += event.content
    elif event.kind in ("tool_completed", "tool_failed"):
        tool_block.output = event.content
    if event.kind in ("tool_approval_requested", "run_suspended"):
        tool_block.status = "awaiting_approval"
        tool_block.output = event.content
    elif event.kind == "tool_approval_resolved":
        tool_block.status = "running"
        tool_block.output = event.content
    elif event.kind == "tool_failed":
        tool_block.status = "failed"
    elif event.kind == "tool_completed":
        tool_block.status = "completed"


def _render_tool_block(tool_block: _ToolBlockState) -> Text:
    event_kind: TUIEventKind = (
        "tool_failed"
        if tool_block.status == "failed"
        else "tool_completed"
        if tool_block.status == "completed"
        else "tool_approval_requested"
        if tool_block.status == "awaiting_approval"
        else "tool_started"
    )
    style = EVENT_STYLES[event_kind]
    if tool_block.name is None:
        raise ValueError("tool block must include a tool name before rendering")
    text = Text()
    text.append(TRANSCRIPT_GUTTER)
    text.append("●", style=SUBTLE_BULLET_STYLE)
    text.append(" ")
    text.append(tool_block.name, style=style)
    text.append(f"  {_tool_summary(tool_block)}", style=style)
    return text


def _render_tool_activity_entries(
    tool_activity: _ToolActivityState,
    expanded_tool_activity_ids: set[str],
) -> list[_StreamRenderable]:
    activity_id = _tool_activity_id(tool_activity)
    expanded = activity_id in expanded_tool_activity_ids
    entries = [
        _StreamRenderable(
            renderable=_render_tool_activity_header(tool_activity, expanded=expanded),
            tool_activity_id=activity_id,
        )
    ]
    if expanded:
        for item in tool_activity.items:
            entries.append(
                _StreamRenderable(renderable=_render_expanded_working_history_item(item))
            )
    return entries


def _render_tool_activity_header(
    tool_activity: _ToolActivityState,
    *,
    expanded: bool,
) -> Text:
    blocks = _working_history_tool_blocks(tool_activity)
    activity_id = _tool_activity_id(tool_activity)
    event_kind = _tool_activity_event_kind(blocks)
    style = EVENT_STYLES[event_kind]
    text = Text()
    text.append(TRANSCRIPT_GUTTER)
    text.append("─", style=EXPAND_MARK_STYLE)
    text.append(" ")
    text.append("[-]" if expanded else "[+]", style=EXPAND_MARK_STYLE)
    text.append(" ")
    text.append("work history", style=f"bold {style}")
    text.append(" · ", style=EXPAND_MARK_STYLE)
    text.append(
        f"{len(blocks)} tool {_pluralize('call', len(blocks))}",
        style=style,
    )
    text.stylize(_tool_activity_click_style(activity_id), 0, len(text))
    return text


def _render_expanded_working_history_item(item: _WorkingHistoryItem) -> RenderableType:
    if item.renderable is not None:
        return _indent_renderable(item.renderable)
    if item.tool_block is None:
        raise ValueError("working history item must include content")
    return _render_expanded_tool_block(item.tool_block)


def _render_expanded_tool_block(tool_block: _ToolBlockState) -> Text:
    base = _render_tool_block(tool_block)
    text = Text()
    text.append("  ")
    text.append(base)
    return text


def _pluralize(word: str, count: int) -> str:
    if count == 1:
        return word
    return f"{word}s"


def _indent_renderable(renderable: RenderableType) -> RenderableType:
    if isinstance(renderable, Text):
        text = Text("  ")
        text.append(renderable)
        return text
    return Group(Text("  "), renderable)


def _tool_activity_id(tool_activity: _ToolActivityState) -> str:
    blocks = _working_history_tool_blocks(tool_activity)
    if blocks:
        return "|".join(block.call_id for block in blocks)
    digest = sha1()
    for item in tool_activity.items:
        if item.renderable is not None and isinstance(item.renderable, Text):
            digest.update(item.renderable.plain.encode("utf-8"))
    return f"working-history:{digest.hexdigest()}"


def _tool_activity_click_style(activity_id: str) -> Style:
    return Style(meta={"tool_activity_id": activity_id})


def _tool_activity_id_from_click(event: Click) -> str:
    if event.style is None:
        return ""
    activity_id = event.style.meta.get("tool_activity_id")
    if type(activity_id) is not str:
        return ""
    return activity_id


def _tool_activity_event_kind(blocks: list[_ToolBlockState]) -> TUIEventKind:
    for block in blocks:
        if block.status == "awaiting_approval":
            return "tool_approval_requested"
    for block in blocks:
        if block.status == "failed":
            return "tool_failed"
    return "tool_completed"


def _working_history_tool_blocks(
    tool_activity: _ToolActivityState,
) -> list[_ToolBlockState]:
    blocks: list[_ToolBlockState] = []
    for item in tool_activity.items:
        if item.tool_block is not None:
            blocks.append(item.tool_block)
    return blocks


def _tool_activity_names_summary(blocks: list[_ToolBlockState]) -> str:
    counts: dict[str, int] = {}
    for block in blocks:
        if block.name is None:
            raise ValueError("tool block must include a tool name before rendering")
        counts[block.name] = counts.get(block.name, 0) + 1
    parts: list[str] = []
    for name, count in counts.items():
        if count == 1:
            parts.append(name)
        else:
            parts.append(f"{name} x{count}")
    return ", ".join(parts)


def _tool_summary(tool_block: _ToolBlockState) -> str:
    if tool_block.status == "failed":
        return "failed"
    if tool_block.status == "completed":
        summary = _tool_output_summary(tool_block.output)
        if summary != "":
            return f"completed - {summary}"
        return "completed"
    if tool_block.status == "awaiting_approval":
        return "waiting for approval"
    return "running"


def _tool_output_summary(output: str) -> str:
    if output == "":
        return ""
    if '"entries":[' in output:
        entry_count = output.count('"name"')
        return f"{entry_count} entries"
    if '"bytes_written":' in output:
        return "file written"
    if '"exit_code":0' in output:
        return "command exited 0"
    if '"exit_code":' in output:
        return "command finished"
    if '"matches":' in output:
        return "search finished"
    return "result ready"


def _render_assistant_block(content: str) -> RenderableType:
    style = EVENT_STYLES["assistant_delta"]
    if _looks_like_markdown(content):
        return Group(
            Text(""),
            Markdown(content, style=style),
        )
    lines = content.splitlines()
    text = Text()
    text.append(TRANSCRIPT_GUTTER)
    if not lines:
        return text
    text.append(lines[0], style=style)
    for line in lines[1:]:
        text.append("\n  ")
        text.append(line, style=style)
    return text


def _render_text_block(
    label: str,
    content: str,
    *,
    event_kind: TUIEventKind,
    markdown: bool = False,
) -> RenderableType:
    style = EVENT_STYLES[event_kind]
    if event_kind in ("thinking_delta", "reasoning_summary"):
        return _render_reasoning_line(label, content, style=style)
    text = Text()
    text.append(TRANSCRIPT_GUTTER)
    text.append("●", style=SUBTLE_BULLET_STYLE)
    text.append(" ")
    text.append(label, style=f"bold {style}")
    if content != "":
        text.append("  ")
    text.append(content, style=style)
    return text


def _render_reasoning_line(label: str, content: str, *, style: str) -> Text:
    text = Text()
    text.append(TRANSCRIPT_GUTTER)
    text.append("*", style=REASONING_MARK_STYLE)
    text.append(" ")
    text.append(label, style=f"bold {style}")
    if content != "":
        text.append("  ")
        text.append(content, style=style)
    return text


def _render_idea_list(items: list[TUIIdeaItem]) -> RenderableType:
    if not items:
        return _render_text_block(
            "ideas",
            "No saved ideas for this workspace.",
            event_kind="idea_list",
        )
    renderables: list[RenderableType] = [_render_idea_header(len(items))]
    for item in items:
        renderables.append(_render_idea_item(item))
    return Group(*renderables)


def _render_idea_header(count: int) -> Text:
    text = Text()
    text.append(TRANSCRIPT_GUTTER)
    text.append("Ideas", style=f"bold {EVENT_STYLES['idea_list']}")
    text.append(f"  {count}", style=SUBTLE_BULLET_STYLE)
    return text


def _render_idea_item(item: TUIIdeaItem) -> Panel:
    title = Text()
    title.append(f"{item.index}. ", style=SUBTLE_BULLET_STYLE)
    title.append(item.title, style="bold #eceff4")
    title.append(f"  {item.created_at}", style="#9aa3b2")
    body = Text()
    body.append(item.body if item.body != "" else " ", style="#d8dee9")
    return Panel(
        body,
        title=title,
        title_align="left",
        border_style="#4c566a",
        padding=(0, 1),
    )


def _render_event(event: TUIEvent) -> RenderableType | None:
    label = EVENT_LABELS[event.kind]
    if event.kind == "user_message":
        return _render_user_message(event.content, label=label, event_kind=event.kind)
    if event.kind == "session_notice":
        return _render_text_block(label, event.content, event_kind=event.kind)
    if event.kind == "idea_list":
        return _render_idea_list(event.idea_items)
    if event.kind == "tool_call_delta":
        return None
    if event.kind == "assistant_delta":
        return _render_assistant_block(event.content)
    if event.kind in ("thinking_delta", "reasoning_summary"):
        return _render_text_block(label, event.content, event_kind=event.kind)
    if event.kind == "llm_retrying":
        return None
    if event.kind in (
        "tool_started",
        "tool_approval_requested",
        "tool_approval_resolved",
        "tool_completed",
        "tool_failed",
        "tool_output",
    ):
        if event.kind == "tool_started":
            return None
        title = _tool_title(label, event)
        return _render_text_block(
            title,
            event.content or event.tool_call_id or "",
            event_kind=event.kind,
        )
    if event.kind == "media":
        return _render_text_block(label, "media segment", event_kind=event.kind)
    if event.kind == "llm_completed":
        if event.content == "":
            return None
        return _render_assistant_block(event.content)
    if event.kind in (
        "step_started",
        "step_completed",
        "run_completed",
        "run_suspended",
    ):
        return None
    return None


def _is_invisible_control_event(event: TUIEvent) -> bool:
    return event.kind in (
        "step_started",
        "step_completed",
        "run_completed",
        "run_suspended",
    )


def _tool_title(label: str, event: TUIEvent) -> str:
    if event.tool_name is not None:
        return f"{label}: {event.tool_name}"
    if event.tool_call_id is not None:
        return f"{label}: {event.tool_call_id}"
    return label


def _render_user_message(
    content: str,
    *,
    label: str,
    event_kind: TUIEventKind,
) -> Table:
    row = Table.grid(expand=True)
    row.add_column(ratio=1, style=PROMPT_BAR_STYLE)
    text = Text()
    text.append("▌ ", style=PROMPT_MARK_STYLE)
    text.append(content, style=PROMPT_BAR_STYLE)
    row.add_row(text, style=PROMPT_BAR_STYLE)
    return row


def _looks_like_markdown(content: str) -> bool:
    for line in content.splitlines():
        if line.startswith(("#", "> ", "- ", "* ", "```")):
            return True
        if line[:2].isdigit() and line[2:4] == ". ":
            return True
    return "`" in content or "**" in content or "__" in content


def _available_stream_width(stream: StreamWidget) -> int:
    widths = (
        stream.scrollable_content_region.width,
        stream.content_size.width,
        stream.size.width,
    )
    for width in widths:
        if width > 4:
            return width - 4
    return 1



from pathlib import Path

import pytest
from ididi import Graph

from aceai.core import AgentBase, ToolExecutor
from aceai.llm import LLMResponse
from aceai.llm.models import LLMMessage, LLMStreamEvent, LLMToolCall


class StubSkillSelectingLLMService:
    def __init__(self) -> None:
        self.calls: list[dict] = []

    async def stream(self, **request):
        self.calls.append(request)
        call_index = len(self.calls)

        if call_index == 1:
            messages: list[LLMMessage] = request["messages"]
            tools = request["tools"]
            system_text = messages[0].content[0]["data"]
            user_text = messages[-1].content[0]["data"]
            tool_names = {tool.name for tool in tools}

            assert "prepare the release" in user_text
            assert "<available_skills>" in system_text
            assert "<name>release</name>" in system_text
            assert "Use for release workflows." in system_text
            assert "skill_view" in tool_names

            yield LLMStreamEvent(
                event_type="response.completed",
                response=LLMResponse(
                    text="loading release skill",
                    tool_calls=[
                        LLMToolCall(
                            name="skill_view",
                            arguments='{"name":"release"}',
                            call_id="skill-call-1",
                        )
                    ],
                ),
            )
            return

        if call_index == 2:
            messages = request["messages"]
            tool_messages = [message for message in messages if message.role == "tool"]
            assert len(tool_messages) == 1
            assert tool_messages[0].name == "skill_view"
            assert "Always run the release checklist." in tool_messages[0].content[0][
                "data"
            ]

            yield LLMStreamEvent(
                event_type="response.completed",
                response=LLMResponse(text="Release skill loaded and applied."),
            )
            return

        raise AssertionError("Unexpected extra LLM call")

    async def complete(self, **request) -> LLMResponse:
        raise AssertionError("AgentBase should not call complete() in streaming mode")


def write_skill(root: Path) -> Path:
    skill_dir = root / "release"
    skill_dir.mkdir(parents=True)
    (skill_dir / "SKILL.md").write_text(
        "\n".join(
            [
                "---",
                "name: release",
                "description: Use for release workflows.",
                "---",
                "# Release",
                "Always run the release checklist.",
            ]
        ),
        encoding="utf-8",
    )
    return skill_dir


@pytest.mark.anyio
async def test_agent_loads_skill_after_model_requests_skill_view(
    tmp_path: Path,
) -> None:
    skills_root = tmp_path / "skills"
    write_skill(skills_root)
    llm_service = StubSkillSelectingLLMService()
    executor = ToolExecutor(Graph(), [])
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,
        executor=executor,
        skill_path=skills_root,
        max_steps=2,
    )

    answer = await agent.ask("please prepare the release")

    assert answer == "Release skill loaded and applied."
    assert len(llm_service.calls) == 2
    assert executor.tools["skill_view"].name == "skill_view"



from pathlib import Path

import pytest

from aceai.agent.ace_agent import ACE_AGENT_SKILL_PATH, build_ace_agent
from aceai.agent.features import default_agent_tools
from aceai.agent.features.tools import read_text_file
from aceai.core import ToolExecutionError, ToolExecutor
from aceai.llm.interface import UNSET


def write_skill(root: Path, name: str, description: str) -> None:
    skill_dir = root / name
    skill_dir.mkdir(parents=True)
    (skill_dir / "SKILL.md").write_text(
        "---\n"
        f"name: {name}\n"
        f"description: {description}\n"
        "---\n"
        f"# {name}\n",
        encoding="utf-8",
    )


def test_default_agent_tools_are_product_capabilities() -> None:
    tools = default_agent_tools()
    tool_names = {tool.name for tool in tools}

    assert tool_names == {
        "list_directory",
        "read_text_file",
        "write_text_file",
        "replace_text_in_file",
        "preview_patch",
        "apply_patch",
        "git_status",
        "git_diff",
        "run_shell_command",
        "search_text",
    }
    approval_policies = {tool.name: tool.metadata.approval_policy for tool in tools}
    assert approval_policies["write_text_file"] == "filesystem_write"
    assert approval_policies["replace_text_in_file"] == "filesystem_write"
    assert approval_policies["apply_patch"] == "filesystem_patch"
    assert approval_policies["run_shell_command"] == "shell_command"
    assert {tool.name for tool in tools if tool.metadata.require_approval} == {
        "write_text_file",
        "replace_text_in_file",
        "apply_patch",
        "run_shell_command",
    }


def test_default_agent_tools_apply_permission_overrides() -> None:
    default_tool_map = {tool.name: tool for tool in default_agent_tools()}
    tools = default_agent_tools(
        {
            "read_text_file": "ask",
            "run_shell_command": "always",
            "apply_patch": "never",
        }
    )
    tool_names = {tool.name for tool in tools}
    approval_required = {tool.name for tool in tools if tool.metadata.require_approval}

    assert "apply_patch" not in tool_names
    assert "read_text_file" in approval_required
    assert "run_shell_command" not in approval_required
    assert default_tool_map["run_shell_command"].metadata.require_approval


def test_build_ace_agent_wires_app_tools_and_project_skills(
    tmp_path, monkeypatch
) -> None:
    monkeypatch.setenv("HOME", str(tmp_path / "home"))
    project = tmp_path / "project"
    write_skill(project / ".agent" / "skills", "aceai-release", "Release workflow.")
    monkeypatch.chdir(project)

    agent = build_ace_agent(api_key="test-key", model="gpt-5.5")

    assert agent.default_model == "gpt-5.5"
    assert agent.max_steps is UNSET
    assert ACE_AGENT_SKILL_PATH == "auto"
    assert isinstance(agent._executor, ToolExecutor)
    assert agent._hosted_tools[0].provider_name == "openai"
    assert agent._hosted_tools[0].native_name == "web_search"
    assert set(agent._executor.tools) >= {
        "list_directory",
        "read_text_file",
        "write_text_file",
        "replace_text_in_file",
        "preview_patch",
        "apply_patch",
        "git_status",
        "git_diff",
        "run_shell_command",
        "search_text",
        "skills_list",
        "skill_view",
    }
    assert set(agent.skill_registry.skills) == {"aceai-release"}


def test_build_ace_agent_supports_deepseek_without_openai_hosted_tools(
    tmp_path, monkeypatch
) -> None:
    monkeypatch.setenv("HOME", str(tmp_path / "home"))

    agent = build_ace_agent(
        provider_name="deepseek",
        api_key="test-key",
        model="deepseek-v4-flash",
    )

    assert agent.default_model == "deepseek-v4-flash"
    assert agent._hosted_tools == []


def test_app_file_tool_reports_missing_file_as_tool_execution_error(tmp_path) -> None:
    missing_path = tmp_path / "missing.py"

    with pytest.raises(ToolExecutionError, match="No such file or directory"):
        read_text_file(path=str(missing_path))



#!/usr/bin/env python3
"""Automate release workflow using GitPython."""
import argparse
import re
import subprocess
import sys
from pathlib import Path

from git import GitCommandError, Repo
from packaging.version import InvalidVersion, Version

from aceai.llm.errors import AceAIValidationError

PROJECT_ROOT = Path(__file__).resolve().parents[1]
VERSION_FILE = PROJECT_ROOT / "aceai" / "__init__.py"


def detect_default_base_branch(repo: Repo) -> str:
    try:
        ref = repo.git.symbolic_ref("refs/remotes/origin/HEAD")
    except GitCommandError:
        return "main"
    return ref.rsplit("/", 1)[-1]


def ensure_local_branch(repo: Repo, branch_name: str) -> None:
    if branch_name in repo.heads:
        return
    origin = repo.remotes.origin
    origin.fetch(branch_name)


def infer_version_from_branch(repo: Repo) -> str:
    if repo.head.is_detached:
        raise SystemExit("Cannot infer version from detached HEAD")
    branch_name = repo.active_branch.name
    match = re.match(r"version/(?P<version>.+)$", branch_name)
    if not match:
        raise SystemExit(
            f"Provide --version or check out version/<x.y.z>; current branch is {branch_name}"
        )
    version = match.group("version")
    try:
        Version(version)
    except InvalidVersion as exc:
        raise SystemExit(f"Invalid version suffix on branch {branch_name}") from exc
    return version


def ensure_release_branch(repo: Repo, base_branch: str, release_branch: str) -> None:
    if repo.head.is_detached:
        raise SystemExit("Repository HEAD is detached. Checkout a branch first.")
    if repo.active_branch.name == release_branch:
        return
    if release_branch in repo.heads:
        repo.git.checkout(release_branch)
        return
    ensure_local_branch(repo, base_branch)
    repo.git.checkout(base_branch)
    repo.git.checkout("-b", release_branch)
    print(f"Created release branch {release_branch} from {base_branch}")


def read_current_version() -> str:
    pattern = re.compile(r"__version__\s*=\s*\"([^\"]+)\"")
    content = VERSION_FILE.read_text()
    match = pattern.search(content)
    if not match:
        raise SystemExit("Could not locate __version__ assignment in aceai/__init__.py")
    return match.group(1)


def write_version(new_version: str) -> None:
    pattern = re.compile(r"__version__\s*=\s*\"([^\"]+)\"")
    content = VERSION_FILE.read_text()
    if not pattern.search(content):
        raise SystemExit("Could not locate __version__ assignment in aceai/__init__.py")
    updated = pattern.sub(f'__version__ = "{new_version}"', content, count=1)
    VERSION_FILE.write_text(updated)
    print(f"Updated __version__ to {new_version}")


def ensure_version_order(
    current: str,
    target: str,
    latest_remote_tag: str | None,
    *,
    skip_update: bool,
) -> tuple[Version, Version]:
    current_v = Version(current)
    target_v = Version(target)
    if skip_update:
        if current_v != target_v:
            raise SystemExit(
                "--skip-version-update set but __version__ does not match target"
            )
    # Compare the version that will actually live on this branch (after a potential
    # version-file update) against the latest remote tag.
    branch_version = current_v if current_v == target_v else target_v
    if latest_remote_tag is not None:
        latest_v = Version(latest_remote_tag)
        if branch_version <= latest_v:
            raise SystemExit(
                f"Local branch version {branch_version} must be greater than latest remote tag v{latest_v}"
            )
    return current_v, target_v


def bump_version(version: Version, increment: str) -> Version:
    if increment == "patch":
        return Version(f"{version.major}.{version.minor}.{version.micro + 1}")
    if increment == "minor":
        return Version(f"{version.major}.{version.minor + 1}.0")
    if increment == "major":
        return Version(f"{version.major + 1}.0.0")
    raise AceAIValidationError(f"Unsupported increment: {increment}")


def stage_and_commit(repo: Repo, version: str) -> None:
    repo.git.add(all=True)
    staged = repo.git.diff("--cached", "--name-only")
    if not staged.strip():
        print("No staged changes; skipping commit.")
        return
    repo.index.commit(f"Release version {version}")
    print("Created release commit")


def merge_into_base(repo: Repo, base_branch: str, source_branch: str) -> None:
    ensure_local_branch(repo, base_branch)
    repo.git.checkout(base_branch)
    repo.git.merge(source_branch)
    print(f"Merged {source_branch} into {base_branch}")


def tag_release(repo: Repo, version: str) -> str:
    tag_name = f"v{version}"
    if any(tag.name == tag_name for tag in repo.tags):
        raise SystemExit(f"Tag {tag_name} already exists")
    repo.create_tag(tag_name, message=f"Release version {version}")
    print(f"Tagged release {tag_name}")
    return tag_name


def push_changes(repo: Repo, base_branch: str, tag_name: str) -> None:
    repo.git.push("--atomic", "origin", base_branch, tag_name)
    print("Pushed base branch and tag to origin")


def read_latest_remote_tag_version(repo: Repo, remote_name: str = "origin") -> str | None:
    """
    Return the highest semantic version from remote tags matching `vX.Y.Z`.

    Uses `git ls-remote --tags` so we don't rely on local tag state.
    """
    try:
        output = repo.git.ls_remote("--tags", remote_name)
    except GitCommandError as exc:
        raise SystemExit(f"Failed to read remote tags from {remote_name}: {exc}") from exc

    pattern = re.compile(r"^refs/tags/v(?P<version>[^\\^]+)$")
    versions: list[Version] = []
    for line in output.splitlines():
        parts = line.split("\t", 1)
        if len(parts) != 2:
            continue
        ref = parts[1]
        if ref.endswith("^{}"):
            continue
        match = pattern.match(ref)
        if not match:
            continue
        versions.append(Version(match.group("version")))

    if not versions:
        return None
    return str(max(versions))


def run_build() -> None:
    subprocess.run(["uv", "build"], cwd=PROJECT_ROOT, check=True)
    print("uv build completed")


def run_tests() -> None:
    subprocess.run([
        "uv",
        "run",
        "--group",
        "dev",
        "pytest",
        "-v",
    ], cwd=PROJECT_ROOT, check=True)
    print("pytest suite passed")


def create_new_branch(repo: Repo, *, base_branch: str | None, increment: str) -> None:
    base = base_branch or detect_default_base_branch(repo)
    ensure_local_branch(repo, base)
    repo.git.checkout(base)
    current_version = Version(read_current_version())
    next_version = bump_version(current_version, increment)
    new_branch_name = f"version/{next_version}"
    if new_branch_name in repo.heads:
        raise SystemExit(f"Branch {new_branch_name} already exists")
    repo.git.checkout("-b", new_branch_name)
    print(f"Created branch {new_branch_name} off {base}")


def delete_branch(repo: Repo, version: str) -> None:
    branch_name = f"version/{version}"
    if branch_name in repo.heads:
        repo.git.branch("-D", branch_name)
        print(f"Deleted local branch {branch_name}")
    try:
        repo.git.push("origin", "--delete", branch_name)
        print(f"Deleted remote branch {branch_name}")
    except GitCommandError:
        print(f"Remote branch {branch_name} was not found; skipping remote delete")


def create_next_version_branch(
    repo: Repo, *, base_branch: str, released_version: str, increment: str = "patch"
) -> None:
    """Create version/<next> for the upcoming release cycle."""
    next_version = bump_version(Version(released_version), increment)
    next_branch = f"version/{next_version}"

    if next_branch in repo.heads:
        print(f"Next-version branch {next_branch} already exists; skipping creation.")
        return

    ensure_local_branch(repo, base_branch)
    repo.git.checkout(base_branch)
    repo.git.checkout("-b", next_branch)
    print(f"Created next-version branch {next_branch} from {base_branch}")


def handle_release(repo: Repo, args: argparse.Namespace) -> None:
    base_branch = args.base_branch or detect_default_base_branch(repo)
    release_version = args.version or infer_version_from_branch(repo)
    release_branch = f"version/{release_version}"

    ensure_release_branch(repo, base_branch, release_branch)

    current_version = read_current_version()
    latest_remote_tag = read_latest_remote_tag_version(repo, "origin")
    ensure_version_order(
        current_version,
        release_version,
        latest_remote_tag,
        skip_update=args.skip_version_update,
    )

    if not args.skip_version_update and current_version != release_version:
        write_version(release_version)

    run_tests()
    stage_and_commit(repo, release_version)
    merge_into_base(repo, base_branch, release_branch)
    tag_name = tag_release(repo, release_version)
    push_changes(repo, base_branch, tag_name)
    run_build()
    create_next_version_branch(
        repo,
        base_branch=base_branch,
        released_version=release_version,
        increment="patch",
    )


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="AceAI release automation CLI")
    subparsers = parser.add_subparsers(dest="command", required=True)

    release_parser = subparsers.add_parser(
        "release", help="Perform a full release (merge, tag, push, build)"
    )
    release_parser.add_argument(
        "--version",
        help="Target semantic version (defaults to version/<x.y.z> suffix from current branch)",
    )
    release_parser.add_argument(
        "--base-branch",
        help="Base branch to merge into (defaults to origin/HEAD or main)",
    )
    release_parser.add_argument(
        "--skip-version-update",
        action="store_true",
        help="Skip editing __version__ (expects file already updated)",
    )

    new_branch_parser = subparsers.add_parser(
        "new-branch", help="Create a new version/<x.y.z> branch from the base branch"
    )
    new_branch_parser.add_argument(
        "--base-branch",
        help="Base branch to branch off (defaults to origin/HEAD or main)",
    )
    new_branch_parser.add_argument(
        "--increment",
        choices=["patch", "minor", "major"],
        default="patch",
        help="Which part of the semantic version to bump (default: patch)",
    )

    delete_parser = subparsers.add_parser(
        "delete-branch", help="Delete a version/<x.y.z> branch locally and remotely"
    )
    delete_parser.add_argument(
        "--version",
        required=True,
        help="Version suffix of the branch to delete (e.g., 0.2.0)",
    )

    return parser


def main() -> None:
    parser = build_parser()
    args = parser.parse_args()
    repo = Repo(PROJECT_ROOT)

    if args.command == "release":
        handle_release(repo, args)
    elif args.command == "new-branch":
        create_new_branch(
            repo,
            base_branch=args.base_branch,
            increment=args.increment,
        )
    elif args.command == "delete-branch":
        delete_branch(repo, args.version)
    else:
        parser.error(f"Unknown command {args.command}")


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)



"""Command input widget for the read-only TUI prototype."""

from textual.events import Click, Key
from textual.message import Message
from textual.widgets import Static, TextArea


class CommandInput(TextArea):
    """Bottom input bar for interactive TUI runs."""

    class Submitted(Message):
        def __init__(self, input: "CommandInput", value: str) -> None:
            self.input = input
            self.value = value
            super().__init__()

    class CompletionRequested(Message):
        def __init__(self, input: "CommandInput") -> None:
            self.input = input
            super().__init__()

    BINDINGS = [
        ("escape", "exit_input_mode", "Exit input"),
    ]

    DEFAULT_CSS = """
    CommandInput {
        height: 3;
        background: #3b4252;
        color: #eceff4;
        border: solid #88c0d0;
    }
    """

    def __init__(self, *, id: str | None = None) -> None:
        super().__init__(
            placeholder="Ask AceAI or type /quit",
            id=id,
            show_line_numbers=False,
            soft_wrap=True,
        )

    @property
    def value(self) -> str:
        return self.text

    @value.setter
    def value(self, text: str) -> None:
        self.load_text(text)
        lines = text.splitlines()
        if not lines:
            self.cursor_location = (0, 0)
            return
        self.cursor_location = (len(lines) - 1, len(lines[-1]))

    def on_key(self, event: Key) -> None:
        if event.key == "shift+enter":
            self.insert("\n")
            event.stop()
            return
        if event.key == "enter":
            self.post_message(self.Submitted(self, self.text))
            event.stop()
            return
        if event.key == "tab" and self.text.startswith("/"):
            self.post_message(self.CompletionRequested(self))
            event.stop()
            return
        if event.key == "escape" and self.app.cancel_active_run():
            event.stop()

    def action_exit_input_mode(self) -> None:
        """Return focus to the main stream so app-level shortcuts work."""
        if self.app.cancel_active_run():
            return
        self.blur()
        self.app.query_one("#stream").focus()


class CommandCompletionWidget(Static):
    """Inline slash-command completion hints."""

    DEFAULT_CSS = """
    CommandCompletionWidget {
        height: auto;
        max-height: 4;
        padding: 0 1;
        background: #2e3440;
        color: #d8dee9;
        border: tall #4c566a;
    }

    CommandCompletionWidget.hidden {
        display: none;
    }
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.display_text = ""

    def show_commands(self, commands: list[str]) -> None:
        if not commands:
            self.add_class("hidden")
            self.display_text = ""
            self.update("")
            return
        self.remove_class("hidden")
        self.display_text = "  ".join(f"/{command}" for command in commands)
        self.update(self.display_text)

    def hide(self) -> None:
        self.add_class("hidden")
        self.display_text = ""
        self.update("")

    @property
    def renderable(self) -> str:
        return self.display_text


class QueuedTurnsWidget(Static):
    """Clickable queued messages shown above the input bar."""

    class Selected(Message):
        def __init__(self, *, index: int) -> None:
            super().__init__()
            self.index = index

    DEFAULT_CSS = """
    QueuedTurnsWidget {
        height: auto;
        max-height: 5;
        padding: 0 1;
        background: #3b4252;
        color: #d8dee9;
        border-top: solid #5e81ac;
        border-bottom: none;
    }

    QueuedTurnsWidget.hidden {
        display: none;
    }
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._questions: tuple[str, ...] = ()
        self.display_text = ""
        self.display = False

    def set_questions(self, questions: tuple[str, ...]) -> None:
        self._questions = questions
        if not questions:
            self.display = False
            self.add_class("hidden")
            self.display_text = ""
            self.update("")
            return
        self.display = True
        self.remove_class("hidden")
        self.display_text = "\n".join(
            ["queued - click a line to steer"]
            + [
                _queued_button_label(index, question)
                for index, question in enumerate(questions)
            ]
        )
        self.update(self.display_text)

    @property
    def renderable(self) -> str:
        return self.display_text

    def on_click(self, event: Click) -> None:
        index = event.y - 1
        if index < 0:
            return
        if index >= len(self._questions):
            return
        event.stop()
        self.post_message(self.Selected(index=index))


def _queued_button_label(index: int, question: str) -> str:
    first_line = question.splitlines()[0] if question.splitlines() else ""
    if len(first_line) > 96:
        first_line = f"{first_line[:93]}..."
    return f"{index + 1}. {first_line}"



"""Static event fixtures for the read-only TUI prototype."""

from aceai.core.events import AgentEvent, AgentEventBuilder
from aceai.llm.models import LLMResponse, LLMSegment, LLMToolCall, LLMToolCallDelta
from aceai.core.models import AgentStep, ToolExecutionResult

from .events import TUIEvent


def static_demo_events() -> list[TUIEvent]:
    builder = AgentEventBuilder(step_index=0, step_id="demo-step-1")
    call = LLMToolCall(
        name="search_docs",
        arguments='{"query":"aceai tui"}',
        call_id="call-search-docs",
    )
    response = LLMResponse(
        text="I found the relevant TUI design notes.",
        tool_calls=[call],
        segments=[
            LLMSegment(
                type="reasoning",
                content="Need the design notes before summarizing implementation steps.",
            )
        ],
    )
    step = AgentStep(llm_response=response, tool_results=[])
    result = ToolExecutionResult(
        call=call,
        output='{"matches":["spec/tui.md","docs/tui.md"]}',
    )
    step.tool_results.append(result)

    agent_events: list[AgentEvent] = [
        builder.llm_started(),
        builder.llm_text_delta(text_delta="Checking the TUI plan "),
        builder.llm_tool_call_delta(
            tool_call_delta=LLMToolCallDelta(
                id=call.call_id,
                arguments_delta='{"query":"aceai tui"}',
            )
        ),
        builder.llm_reasoning(segment=response.segments[0]),
        builder.llm_completed(step=step),
        builder.tool_started(tool_call=call),
        builder.tool_completed(tool_call=call, tool_result=result),
        builder.step_completed(step=step),
        builder.run_completed(
            step=step,
            final_answer="Static TUI prototype is ready to inspect.",
        ),
    ]
    return [TUIEvent.from_agent_event(event) for event in agent_events]



import pytest

from aceai.llm.deepseek import DEEPSEEK_BASE_URL, DeepSeek
from aceai.llm.models import LLMMessage, LLMToolCall, LLMToolCallMessage


class RecordingCompletions:
    def __init__(self) -> None:
        self.kwargs: dict[str, object] | None = None

    async def create(self, **kwargs):
        self.kwargs = kwargs
        return ChatCompletionResponse()


class RecordingClient:
    def __init__(self) -> None:
        self.chat = ChatResource()


class ChatResource:
    def __init__(self) -> None:
        self.completions = RecordingCompletions()


class ChatCompletionResponse:
    id = "chatcmpl-test"
    model = "deepseek-v4-flash"
    usage = None
    choices = []

    def __init__(self) -> None:
        self.choices = [ChatChoice()]


class ChatChoice:
    finish_reason = "stop"
    message = None

    def __init__(self) -> None:
        self.message = ChatMessage()


class ChatMessage:
    content = "hello from deepseek"
    tool_calls = None
    model_extra = {"reasoning_content": "thinking about hello"}


class StreamingCompletions:
    def __init__(self) -> None:
        self.kwargs: dict[str, object] | None = None

    async def create(self, **kwargs):
        self.kwargs = kwargs
        return StreamingResponse()


class StreamingClient:
    def __init__(self) -> None:
        self.chat = StreamingChatResource()


class StreamingChatResource:
    def __init__(self) -> None:
        self.completions = StreamingCompletions()


class StreamingResponse:
    def __aiter__(self):
        return self._events()

    async def _events(self):
        yield StreamChunk(choices=[StreamChoice(delta=StreamDelta(content="hi"))])
        yield StreamChunk(
            choices=[],
            usage=StreamUsage(
                prompt_tokens=10,
                completion_tokens=3,
                total_tokens=13,
                prompt_tokens_details=StreamPromptTokenDetails(cached_tokens=4),
            ),
        )


class StreamChunk:
    def __init__(self, *, choices, usage=None) -> None:
        self.choices = choices
        self.usage = usage


class StreamChoice:
    def __init__(self, *, delta) -> None:
        self.delta = delta


class StreamDelta:
    tool_calls = None
    model_extra = None

    def __init__(self, *, content=None) -> None:
        self.content = content


class StreamUsage:
    def __init__(
        self,
        *,
        prompt_tokens,
        completion_tokens,
        total_tokens,
        prompt_tokens_details,
    ) -> None:
        self.prompt_tokens = prompt_tokens
        self.completion_tokens = completion_tokens
        self.total_tokens = total_tokens
        self.prompt_tokens_details = prompt_tokens_details
        self.prompt_cache_hit_tokens = prompt_tokens_details.cached_tokens
        self.prompt_cache_miss_tokens = prompt_tokens - prompt_tokens_details.cached_tokens


class StreamPromptTokenDetails:
    def __init__(self, *, cached_tokens) -> None:
        self.cached_tokens = cached_tokens


def test_deepseek_provider_uses_openai_compatible_base_url() -> None:
    provider = DeepSeek(
        api_key="test-key",
        default_meta={"model": "deepseek-v4-flash"},
    )

    assert str(provider._client.base_url) == DEEPSEEK_BASE_URL
    assert provider._provider_name == "deepseek"


@pytest.mark.anyio
async def test_deepseek_complete_uses_chat_completions() -> None:
    provider = DeepSeek(
        api_key="test-key",
        default_meta={"model": "deepseek-v4-flash"},
    )
    client = RecordingClient()
    provider._client = client

    response = await provider.complete(
        {
            "messages": [
                LLMMessage.build(role="user", content="hello"),
            ],
        }
    )

    assert client.chat.completions.kwargs == {
        "model": "deepseek-v4-flash",
        "messages": [{"role": "user", "content": "hello"}],
    }
    assert response.text == "hello from deepseek"
    assert response.reasoning_content == "thinking about hello"
    assert response.provider_meta[0].provider_name == "deepseek"


def test_deepseek_formats_reasoning_content_for_tool_sub_turn() -> None:
    provider = DeepSeek(
        api_key="test-key",
        default_meta={"model": "deepseek-v4-flash"},
    )
    call = LLMToolCall(
        name="lookup",
        arguments='{"q":"aceai"}',
        call_id="call-1",
    )
    messages = provider._format_messages_for_chat(
        [
            LLMToolCallMessage.from_content(
                content="",
                tool_calls=[call],
                reasoning_content="need to look this up",
            )
        ]
    )

    assert messages == [
        {
            "role": "assistant",
            "content": "",
            "reasoning_content": "need to look this up",
            "tool_calls": [
                {
                    "id": "call-1",
                    "type": "function",
                    "function": {
                        "name": "lookup",
                        "arguments": '{"q":"aceai"}',
                    },
                }
            ],
        }
    ]


@pytest.mark.anyio
async def test_deepseek_stream_requests_and_maps_usage() -> None:
    provider = DeepSeek(
        api_key="test-key",
        default_meta={"model": "deepseek-v4-flash"},
    )
    client = StreamingClient()
    provider._client = client

    events = [
        event
        async for event in provider.stream(
            {
                "messages": [
                    LLMMessage.build(role="user", content="hello"),
                ],
            }
        )
    ]

    assert client.chat.completions.kwargs is not None
    assert client.chat.completions.kwargs["stream"] is True
    assert client.chat.completions.kwargs["stream_options"] == {"include_usage": True}
    final_response = events[-1].response
    assert final_response.usage.input_tokens == 10
    assert final_response.usage.cached_input_tokens == 4
    assert final_response.usage.cache_miss_input_tokens == 6
    assert final_response.usage.input_cache_hit_rate == 0.4
    assert final_response.usage.output_tokens == 3
    assert final_response.usage.total_tokens == 13



from dataclasses import dataclass
from inspect import Parameter, Signature
from typing import Annotated as Annotated
from typing import Any, Callable, TypedDict, Unpack, get_args, get_origin

from ididi import DependentNode, NodeMeta, is_provided
from ididi.utils.typing_utils import flatten_annotated
from msgspec import Meta, Struct, defstruct

from aceai.llm.errors import AceAIConfigurationError, UnannotatedToolParamError
from aceai.llm.interface import MISSING, JsonSchema, Maybe, is_present
from aceai.core.tools.schema_generator import inline_schema


class ParamConstraint(TypedDict, total=False):
    gt: int | float
    ge: int | float
    lt: int | float
    le: int | float
    multiple_of: int | float
    pattern: str
    min_length: int
    max_length: int
    tz: bool
    extra: dict[str, Any]


@dataclass(frozen=True, kw_only=True, slots=True)
class ParamSpec[T]:
    alias: Maybe[str] = MISSING
    description: str
    required: Maybe[bool] = MISSING
    annotation: Maybe[type[T]] = MISSING
    title: Maybe[str] = MISSING
    examples: list[str]
    extra_json_schema: JsonSchema
    constraint: ParamConstraint


def spec[T](
    description: str = "",
    alias: Maybe[str] = MISSING,
    required: Maybe[bool] = MISSING,
    annotation: Maybe[type[T]] = MISSING,
    title: Maybe[str] = MISSING,
    examples: Maybe[list[Any]] = MISSING,
    extra_json_schema: Maybe[JsonSchema] = MISSING,
    **constraint: Unpack[ParamConstraint],
) -> ParamSpec[T]:
    """
    Args:
        alias: The name of the parameter in the tool schema.
        description: A description of the parameter.
        required: Whether the parameter is required, if not specified, defaults to whether a default value for the param is provided.
        annotation: The type annotation override of the parameter. default is to use the annotated type.
        constraint: Additional constraints for the parameter, see `ParamConstraint` for details.
    """
    if not is_present(examples):
        examples = []
    if not is_present(extra_json_schema):
        extra_json_schema = {}
    return ParamSpec(
        alias=alias,
        description=description,
        required=required,
        annotation=annotation,
        title=title,
        examples=examples,
        extra_json_schema=extra_json_schema,
        constraint=constraint,
    )


def get_param_meta(param: Parameter) -> list[Any]:
    if not param.annotation:
        raise UnannotatedToolParamError(
            f"Tool parameter {param.name!r} must use typing.Annotated with spec or dep metadata"
        )

    t_origin = get_origin(param.annotation)
    if t_origin is not Annotated:
        raise UnannotatedToolParamError(
            f"Tool parameter {param.name!r} must use typing.Annotated with spec or dep metadata"
        )

    param_meta = flatten_annotated(param.annotation)
    return param_meta


def get_param_spec(param_metas: list[Any] | None) -> ParamSpec[Any] | None:
    if not param_metas:
        return None

    for meta in param_metas:
        if isinstance(meta, ParamSpec):
            return meta


def get_dep_node(param_metas: list[Any], param_type: Any) -> DependentNode | None:
    for meta in param_metas:
        if isinstance(meta, NodeMeta):
            if meta.ignore is True:
                return None
            dep = meta.factory if is_provided(meta.factory) else param_type
            return DependentNode.from_node(dep)
    return None


@dataclass(frozen=True, kw_only=True, slots=True)
class ToolParam[T]:
    name: str
    alias: str
    required: bool
    type_: type[T]
    annotation: Any
    default: Maybe[T] = MISSING
    schema: JsonSchema

    @classmethod
    def from_param(
        cls,
        param: Parameter,
        param_spec: ParamSpec[T],
    ) -> "ToolParam[T]":
        default = MISSING if param.default is Parameter.empty else param.default
        alias = param_spec.alias if is_present(param_spec.alias) else param.name

        annotation = (
            param_spec.annotation
            if is_present(param_spec.annotation)
            else param.annotation
        )
        param_type = get_args(annotation)[0]
        required = (
            param_spec.required
            if is_present(param_spec.required)
            else default is MISSING
        )

        param_schema = inline_schema(param_type, default)
        param_schema["description"] = param_spec.description

        if param_spec.examples:
            param_schema["examples"] = param_spec.examples
        if param_spec.extra_json_schema:
            param_schema.update(param_spec.extra_json_schema)
        if param_spec.constraint:
            annotation = Annotated[param_type, Meta(**param_spec.constraint)]

        return cls(
            name=param.name,
            alias=alias,
            type_=param_type,
            annotation=annotation,
            required=required,
            default=default,
            schema=param_schema,
        )


@dataclass(kw_only=True, slots=True)
class ToolSignature:
    params: dict[str, ToolParam[Any]]
    dep_nodes: dict[str, Callable[..., Any]]
    return_type: Maybe[type]

    def generate_params_schema(self) -> dict[str, Any]:
        properties: dict[str, Any] = {}
        required: list[str] = []
        for param in self.params.values():
            properties[param.alias] = param.schema
            if param.required:
                required.append(param.alias)

        parameters: dict[str, Any] = {
            "type": "object",
            "properties": properties,
            "additionalProperties": False,
        }
        if required:
            parameters["required"] = required

        return parameters

    @property
    def virtual_struct(self) -> type[Struct]:
        fields = []

        for param in self.params.values():
            field_tuple = (param.alias, param.annotation)
            if is_present(param.default):
                field_tuple += (param.default,)
            fields.append(field_tuple)

        typed_struct = defstruct(f"ToolParams", fields)
        return typed_struct

    @classmethod
    def from_signature(cls, func_sig: Signature) -> "ToolSignature":
        params: dict[str, ToolParam[Any]] = {}
        dep_nodes: dict[str, Callable[..., Any]] = {}
        for param in func_sig.parameters.values():
            param_metas = get_param_meta(param)
            if not param_metas:
                continue

            param_spec = get_param_spec(param_metas)
            if param_spec:
                tool_param = ToolParam[Any].from_param(param, param_spec)
                params[param.name] = tool_param

            param_type = get_args(param.annotation)[0]
            dep_node = get_dep_node(param_metas, param_type)
            if dep_node:
                if param.name in params:
                    raise AceAIConfigurationError(
                        f"Parameter {param.name!r} is defined as both ToolParam and DependentNode"
                    )
                dep_nodes[param.name] = dep_node.factory

        return_type = (
            func_sig.return_annotation
            if func_sig.return_annotation is not Signature.empty
            else MISSING
        )

        return cls(params=params, dep_nodes=dep_nodes, return_type=return_type)



from aceai.agent.ideas import IdeaStore


def test_idea_store_appends_markdown_and_lists_current_workspace(tmp_path) -> None:
    workspace = tmp_path / "repo"
    workspace.mkdir()
    other_workspace = tmp_path / "other"
    other_workspace.mkdir()
    ideas_path = tmp_path / "ideas.md"
    store = IdeaStore(ideas_path)

    first = store.capture("fix resume default session", workspace=workspace)
    store.capture("hidden other repo idea", workspace=other_workspace)
    second = store.capture("add idea command", workspace=workspace)

    markdown = ideas_path.read_text(encoding="utf-8")
    assert markdown.startswith("# AceAI Ideas\n\n")
    assert f"## {first.created_at.isoformat()}" in markdown
    assert "workspace: " + str(workspace.resolve()) in markdown
    assert "source_session_id: \n\nfix resume default session" in markdown

    ideas = store.list_recent(workspace=workspace)

    assert [idea.created_at for idea in ideas] == [second.created_at, first.created_at]
    assert [idea.content for idea in ideas] == [
        "add idea command",
        "fix resume default session",
    ]


def test_idea_store_searches_by_keyword_within_workspace(tmp_path) -> None:
    workspace = tmp_path / "repo"
    workspace.mkdir()
    store = IdeaStore(tmp_path / "ideas.md")
    store.capture("fix resume default session", workspace=workspace)
    store.capture("add idea command", workspace=workspace)

    ideas = store.search("resume", workspace=workspace)

    assert [idea.content for idea in ideas] == ["fix resume default session"]


def test_idea_store_deletes_recent_idea_by_one_based_workspace_index(tmp_path) -> None:
    workspace = tmp_path / "repo"
    workspace.mkdir()
    store = IdeaStore(tmp_path / "ideas.md")
    first = store.capture("first idea", workspace=workspace)
    store.capture("second idea", workspace=workspace)

    deleted = store.delete_recent(1, workspace=workspace)

    assert deleted.content == "second idea"
    assert [idea.content for idea in store.list_recent(workspace=workspace)] == [
        first.content
    ]
    assert "second idea" not in (tmp_path / "ideas.md").read_text(encoding="utf-8")



"""Read-only Textual application for AceAI event streams."""

from textual.app import App, ComposeResult
from textual.containers import Horizontal
from textual.events import Key
from textual.timer import Timer
from textual.widgets import Footer, Header, TextArea

from aceai.core.events import AgentEvent

from aceai.agent.ideas import IdeaStore
from aceai.agent.session import SessionRecorder, SessionStore

from aceai.agent.cost import format_usd
from .events import TUIEvent
from .metadata import MetadataScreen, MetadataSection
from .session_adapter import tui_event_to_session_event
from .session_display import session_display_title
from .session_replay import event_log_to_tui_events
from .setup import SessionSelectScreen
from .state import TUIRunState, apply_tui_event, initial_state, reduce_events, select_event
from .trajectory import TrajectoryScreen
from .widgets import (
    ApprovalWidget,
    CommandCompletionWidget,
    CommandInput,
    DetailWidget,
    QueuedTurnsWidget,
    StatusBarWidget,
    StreamWidget,
)

STREAM_DELTA_REFRESH_CHARS = 32
STREAM_DELTA_REFRESH_SECONDS = 0.02


class AceAITUI(App[None]):
    """Read-only TUI prototype backed by normalized TUI events."""

    CSS = """
    Screen {
        layout: vertical;
        background: #2e3440;
        color: #e5e9f0;
    }

    Header {
        background: #3b4252;
        color: #eceff4;
        text-style: bold;
    }

    #main {
        height: 1fr;
    }

    #stream {
        min-width: 40;
    }

    DetailWidget.collapsed {
        display: none;
    }

    Footer {
        background: #3b4252;
        color: #eceff4;
    }

    """

    BINDINGS = [
        ("q", "quit", "Quit"),
        ("ctrl+c", "quit", "Quit"),
        ("d", "toggle_debug_mode", "Debug"),
        ("c", "config", "Config"),
        ("t", "trajectory", "Trajectory"),
        ("i", "metadata", "Info"),
        ("s", "session_switcher", "Sessions"),
    ]

    def __init__(
        self,
        events: list[TUIEvent] | None = None,
        *,
        model: str | None = None,
        session_recorder: SessionRecorder | None = None,
        session_id: str | None = None,
        idea_store: IdeaStore | None = None,
        record_events: bool = True,
    ) -> None:
        super().__init__()
        self._events = list(events or [])
        self._state: TUIRunState = initial_state()
        self._status_model = model
        self._session_recorder = session_recorder
        self._session_id = session_id
        self._idea_store = idea_store or IdeaStore()
        self._record_events = record_events
        self._pending_stream_delta_chars = 0
        self._pending_stream_delta: TUIEvent | None = None
        self._pending_stream_delta_timer: Timer | None = None
        self.title = "AceAI" if session_id is None else f"AceAI {session_id}"

    def compose(self) -> ComposeResult:
        yield Header(show_clock=True)
        with Horizontal(id="main"):
            yield StreamWidget(id="stream")
            yield DetailWidget(id="detail", classes="collapsed")
        yield ApprovalWidget(id="approval", classes="collapsed")
        yield StatusBarWidget(id="status")
        yield CommandCompletionWidget(id="command-completions", classes="hidden")
        yield QueuedTurnsWidget(id="queued-turns", classes="hidden")
        yield CommandInput(id="input")
        yield Footer()

    def on_mount(self) -> None:
        self.load_events(self._events)
        self.query_one(StreamWidget).focus()

    def on_key(self, event: Key) -> None:
        if event.key == "escape" and self.cancel_active_run():
            event.stop()
            return
        if event.key != "enter":
            return
        command_input = self.query_one(CommandInput)
        if command_input.has_focus:
            return
        command_input.focus()
        event.stop()

    def load_events(self, events: list[TUIEvent]) -> None:
        self._clear_pending_stream_delta()
        if self.is_mounted:
            self.clear_approval_request()
        self._state = reduce_events(events)
        self._refresh_widgets()

    def show_sessions(self) -> None:
        store = self._session_store()
        sessions = store.list_sessions()
        if not sessions:
            self.append_event(TUIEvent.session_notice("No sessions found."))
            return
        total_cost = store.total_cost_usd()
        lines = [f"Total cost: {format_usd(total_cost)}", "", "Sessions:"]
        for session in sessions:
            marker = "*" if session.session_id == self._session_id else "-"
            lines.append(
                f"{marker} {session.session_id}  "
                f"{session_display_title(session.title)}  {session.updated_at}"
            )
        lines.append("")
        lines.append("Use /resume <session_id> to switch sessions.")
        self.append_event(TUIEvent.session_notice("\n".join(lines)))

    def open_session_selector(self) -> None:
        store = self._session_store()
        sessions = store.list_sessions()
        if not sessions:
            self.append_event(TUIEvent.session_notice("No sessions found."))
            return
        self.push_screen(
            SessionSelectScreen(
                store=store,
                sessions=sessions,
                current_session_id=self._session_id,
            ),
            self._handle_session_selection,
        )

    def _handle_session_selection(self, session_id: str | None) -> None:
        if session_id is None:
            return
        self.switch_session(session_id)

    def switch_session(self, session_id: str) -> None:
        if session_id == self._session_id:
            return
        store = self._session_store()
        try:
            metadata = store.get_session(session_id)
        except KeyError:
            self.notify_session(f"Session not found: {session_id}")
            return
        if self._session_recorder is not None:
            self._session_recorder.finalize()
        self._session_recorder = SessionRecorder(store, metadata.session_id)
        self._session_id = metadata.session_id
        self.title = f"AceAI {metadata.session_id}"
        event_log = store.load_event_log(metadata.session_id)
        self.load_events(event_log_to_tui_events(event_log))
        self.notify_session(f"Resumed session {metadata.session_id}")

    def ensure_session(self) -> None:
        if self._session_recorder is not None:
            return
        store = SessionStore()
        metadata = store.create_session()
        self._session_recorder = SessionRecorder(store, metadata.session_id)
        self._session_id = metadata.session_id
        self.title = f"AceAI {metadata.session_id}"

    def _session_store(self) -> SessionStore:
        if self._session_recorder is not None:
            return self._session_recorder.store
        return SessionStore()

    def append_event(self, event: TUIEvent) -> None:
        if self._should_buffer_stream_delta(event):
            return
        self._flush_pending_stream_delta()
        self._append_event_to_state(event)
        self._refresh_widgets()

    def _append_event_to_state(self, event: TUIEvent) -> None:
        self._state = apply_tui_event(self._state, event)
        if self._record_events and self._session_recorder is not None:
            self._session_recorder.record(tui_event_to_session_event(event))

    def append_agent_event(self, event: AgentEvent) -> None:
        tui_event = TUIEvent.from_agent_event(event)
        if tui_event.kind == "llm_retrying":
            self.notify(
                tui_event.content,
                title="Retrying LLM",
                severity="warning",
                timeout=3.0,
            )
            return
        self.append_event(tui_event)

    def notify_session(self, content: str) -> None:
        self.notify(content, title="AceAI", severity="information", timeout=3.0)

    def cancel_active_run(self) -> bool:
        return False

    def set_status_model(self, model: str | None) -> None:
        self._status_model = model
        if self.is_mounted:
            self.query_one(StatusBarWidget).set_status(
                model=self._status_model,
                status=self._state.status,
                usage=self._state.usage,
            )

    def show_approval_request(self, request) -> None:
        self.query_one(ApprovalWidget).show_request(request)

    def clear_approval_request(self) -> None:
        self.query_one(ApprovalWidget).clear_request()

    def exit_command_input(self, command_input: CommandInput) -> None:
        command_input.value = ""
        self._hide_command_completions()
        command_input.blur()
        self._focus_message_panel()

    def command_names(self) -> tuple[str, ...]:
        return ()

    def on_text_area_changed(self, event: TextArea.Changed) -> None:
        if not isinstance(event.text_area, CommandInput):
            return
        self._refresh_command_completions(event.text_area.value)

    def on_command_input_completion_requested(
        self,
        event: CommandInput.CompletionRequested,
    ) -> None:
        matches = self._matching_command_names(event.input.value)
        if not matches:
            return
        event.input.value = f"/{matches[0]} "
        self._refresh_command_completions(event.input.value)

    def _refresh_command_completions(self, value: str) -> None:
        completions = self._command_completion_widget()
        if completions is None:
            return
        matches = self._matching_command_names(value)
        if not matches:
            completions.hide()
            return
        completions.show_commands(list(matches))

    def _matching_command_names(self, value: str) -> tuple[str, ...]:
        if not value.startswith("/"):
            return ()
        body = value.removeprefix("/")
        if " " in body or "\n" in body:
            return ()
        return tuple(name for name in self.command_names() if name.startswith(body))

    def _hide_command_completions(self) -> None:
        completions = self._command_completion_widget()
        if completions is None:
            return
        completions.hide()

    def _command_completion_widget(self) -> CommandCompletionWidget | None:
        matches = list(self.query(CommandCompletionWidget))
        if not matches:
            return None
        return matches[0]

    def action_toggle_debug_mode(self) -> None:
        stream = self.query_one(StreamWidget)
        selected_event_id = stream.set_debug_mode(not stream.debug_mode)
        stream.focus()
        if selected_event_id is None:
            self.query_one(DetailWidget).add_class("collapsed")
            return
        self._select_debug_event(selected_event_id)

    def action_config(self) -> None:
        self.append_event(
            TUIEvent.session_notice("Configuration is only available in live TUI runs.")
        )

    def action_session_switcher(self) -> None:
        self.open_session_selector()

    def action_metadata(self) -> None:
        self.open_metadata_screen()

    def action_trajectory(self) -> None:
        self.open_trajectory_screen()

    def open_metadata_screen(self) -> None:
        self.push_screen(MetadataScreen(self._metadata_sections()))

    def open_trajectory_screen(self) -> None:
        self.push_screen(TrajectoryScreen(self._state.events))

    def _metadata_sections(self) -> list[MetadataSection]:
        usage = self._state.usage
        lines = [
            f"session: {self._session_id or '-'}",
            f"model: {self._status_model or 'unconfigured'}",
            f"status: {self._state.status}",
            f"events: {len(self._state.events)}",
        ]
        cost_lines = [
            f"context: {_format_tokens(usage.current_context_tokens)}",
            f"session tokens: {_format_tokens(usage.session_total_tokens)}",
            f"input: {_format_tokens(usage.session_input_tokens)}",
            f"cached input: {_format_tokens(usage.session_cached_input_tokens)}",
            f"output: {_format_tokens(usage.session_output_tokens)}",
            f"session cost: {format_usd(usage.session_cost_usd)}",
        ]
        return [
            MetadataSection(title="Runtime", lines=lines),
            MetadataSection(title="Usage", lines=cost_lines),
        ]

    def on_stream_widget_event_selected(
        self,
        event: StreamWidget.EventSelected,
    ) -> None:
        self._select_debug_event(event.event_id)
        event.stop()

    def _select_debug_event(self, event_id: str) -> None:
        self._state = select_event(self._state, event_id)
        self.query_one(DetailWidget).remove_class("collapsed")
        self.query_one(DetailWidget).set_state(self._state)

    def _refresh_widgets(self) -> None:
        self.query_one(StreamWidget).set_state(self._state)
        self.query_one(DetailWidget).set_state(self._state)
        self.query_one(StatusBarWidget).set_status(
            model=self._status_model,
            status=self._state.status,
            usage=self._state.usage,
        )
        command_input = self.query_one(CommandInput)
        if self._state.status == "suspended":
            command_input.placeholder = "Choose Approve or Reject"
        else:
            command_input.placeholder = "Ask AceAI or type /quit"

    def _focus_message_panel(self) -> None:
        self.query_one(StreamWidget).focus()

    def _should_buffer_stream_delta(self, event: TUIEvent) -> bool:
        if event.kind not in ("assistant_delta", "thinking_delta"):
            return False
        if self._pending_stream_delta is not None and not _same_stream_delta(
            self._pending_stream_delta,
            event,
        ):
            self._flush_pending_stream_delta()
        if self._pending_stream_delta is None and not _same_as_last_stream_delta(
            self._state.events,
            event,
        ):
            self._append_event_to_state(event)
            self._refresh_widgets()
            return True
        self._pending_stream_delta_chars += len(event.content)
        self._pending_stream_delta = _merge_pending_stream_delta(
            self._pending_stream_delta,
            event,
        )
        if self._pending_stream_delta_chars < STREAM_DELTA_REFRESH_CHARS:
            self._ensure_pending_stream_delta_timer()
            return True
        self._flush_pending_stream_delta()
        self._refresh_widgets()
        return True

    def _ensure_pending_stream_delta_timer(self) -> None:
        if not self.is_mounted:
            return
        if self._pending_stream_delta_timer is not None:
            return
        self._pending_stream_delta_timer = self.set_timer(
            STREAM_DELTA_REFRESH_SECONDS,
            self._flush_pending_stream_delta_from_timer,
            name="stream-delta-refresh",
        )

    def _flush_pending_stream_delta_from_timer(self) -> None:
        if self._flush_pending_stream_delta():
            self._refresh_widgets()

    def _flush_pending_stream_delta(self) -> bool:
        if self._pending_stream_delta is None:
            self._clear_pending_stream_delta_timer()
            return False
        pending = self._pending_stream_delta
        self._clear_pending_stream_delta()
        self._append_event_to_state(pending)
        return True

    def _clear_pending_stream_delta(self) -> None:
        self._pending_stream_delta_chars = 0
        self._pending_stream_delta = None
        self._clear_pending_stream_delta_timer()

    def _clear_pending_stream_delta_timer(self) -> None:
        timer = self._pending_stream_delta_timer
        self._pending_stream_delta_timer = None
        if timer is not None:
            timer.stop()

    def on_unmount(self) -> None:
        self._flush_pending_stream_delta()
        if self._record_events and self._session_recorder is not None:
            self._session_recorder.finalize()


def run_static_tui(events: list[TUIEvent]) -> None:
    AceAITUI(events).run()


def _merge_pending_stream_delta(previous: TUIEvent | None, event: TUIEvent) -> TUIEvent:
    if previous is None:
        return event
    if not _same_stream_delta(previous, event):
        raise ValueError("pending stream delta changed shape before flush")
    return TUIEvent(
        kind=previous.kind,
        step_index=previous.step_index,
        step_id=previous.step_id,
        title=previous.title,
        raw_event=event.raw_event,
        event_id=previous.event_id,
        content=previous.content + event.content,
        tool_name=previous.tool_name,
        tool_call_id=previous.tool_call_id,
        tool_call=previous.tool_call,
        tool_calls=previous.tool_calls,
        tool_call_delta=previous.tool_call_delta,
        tool_result=previous.tool_result,
        segment=event.segment,
        usage=event.usage,
        cost=event.cost,
        error=event.error,
    )


def _same_stream_delta(previous: TUIEvent, event: TUIEvent) -> bool:
    return (
        previous.kind == event.kind
        and previous.step_id == event.step_id
        and previous.step_index == event.step_index
        and previous.tool_call_id == event.tool_call_id
    )


def _same_as_last_stream_delta(events: list[TUIEvent], event: TUIEvent) -> bool:
    if not events:
        return False
    return _same_stream_delta(events[-1], event)


def _format_tokens(value: int | None) -> str:
    if value is None:
        return "-"
    return f"{value:,}"



"""Provider and model catalog loaded from package data."""

from functools import cache
from importlib import resources

import yaml
from typing_extensions import Self

from aceai.llm.interface import Record


CATALOG_RESOURCE = "provider_catalog.yaml"


class ModelTokenPrice(Record, kw_only=True):
    model: str
    input_usd_per_million: float
    cached_input_usd_per_million: float
    output_usd_per_million: float

    @classmethod
    def from_raw(cls, model_id: str, raw: object) -> Self:
        if type(raw) is not dict:
            raise TypeError("Provider catalog pricing must be a mapping")
        input_price = raw["input_usd_per_million"]
        cached_input_price = raw["cached_input_usd_per_million"]
        output_price = raw["output_usd_per_million"]
        if type(input_price) is not float:
            raise TypeError("Provider catalog input_usd_per_million must be float")
        if type(cached_input_price) is not float:
            raise TypeError("Provider catalog cached_input_usd_per_million must be float")
        if type(output_price) is not float:
            raise TypeError("Provider catalog output_usd_per_million must be float")
        return cls(
            model=model_id,
            input_usd_per_million=input_price,
            cached_input_usd_per_million=cached_input_price,
            output_usd_per_million=output_price,
        )


class ModelCatalogEntry(Record, kw_only=True):
    id: str
    label: str
    pricing: ModelTokenPrice | None = None

    @classmethod
    def from_raw(cls, raw: object) -> Self:
        if type(raw) is not dict:
            raise TypeError("Provider catalog model must be a mapping")
        model_id = raw["id"]
        label = raw["label"]
        if type(model_id) is not str:
            raise TypeError("Provider catalog model id must be str")
        if type(label) is not str:
            raise TypeError("Provider catalog model label must be str")
        pricing = None
        if "pricing" in raw:
            pricing = ModelTokenPrice.from_raw(model_id, raw["pricing"])
        return cls(id=model_id, label=label, pricing=pricing)


class ProviderCatalogEntry(Record, kw_only=True):
    name: str
    label: str
    api_key_env: str
    default_model: str
    stale_default_models: tuple[str, ...]
    models: tuple[ModelCatalogEntry, ...]

    @classmethod
    def from_raw(cls, raw: object) -> Self:
        if type(raw) is not dict:
            raise TypeError("Provider catalog provider must be a mapping")
        name = raw["name"]
        label = raw["label"]
        api_key_env_value = raw["api_key_env"]
        default_model_value = raw["default_model"]
        stale_default_model_values = raw["stale_default_models"]
        models = raw["models"]
        if type(name) is not str:
            raise TypeError("Provider catalog provider name must be str")
        if type(label) is not str:
            raise TypeError("Provider catalog provider label must be str")
        if type(api_key_env_value) is not str:
            raise TypeError("Provider catalog api_key_env must be str")
        if type(default_model_value) is not str:
            raise TypeError("Provider catalog default_model must be str")
        if type(stale_default_model_values) is not list:
            raise TypeError("Provider catalog stale_default_models must be a list")
        if type(models) is not list:
            raise TypeError("Provider catalog models must be a list")
        parsed_stale_models = tuple(
            _require_str(model, "Provider catalog stale model must be str")
            for model in stale_default_model_values
        )
        parsed_models = tuple(ModelCatalogEntry.from_raw(model) for model in models)
        return cls(
            name=name,
            label=label,
            api_key_env=api_key_env_value,
            default_model=default_model_value,
            stale_default_models=parsed_stale_models,
            models=parsed_models,
        )


class ProviderCatalog(Record, kw_only=True):
    providers: tuple[ProviderCatalogEntry, ...]
    pricing_source: str

    @classmethod
    def from_raw(cls, raw: object) -> Self:
        if type(raw) is not dict:
            raise TypeError("Provider catalog must be a mapping")
        providers = raw["providers"]
        pricing_source_value = raw["pricing_source"]
        if type(providers) is not list:
            raise TypeError("Provider catalog providers must be a list")
        if type(pricing_source_value) is not str:
            raise TypeError("Provider catalog pricing_source must be str")
        return cls(
            providers=tuple(
                ProviderCatalogEntry.from_raw(provider) for provider in providers
            ),
            pricing_source=pricing_source_value,
        )


@cache
def load_provider_catalog() -> ProviderCatalog:
    text = resources.files("aceai.agent").joinpath(CATALOG_RESOURCE).read_text(
        encoding="utf-8"
    )
    raw = yaml.safe_load(text)
    return ProviderCatalog.from_raw(raw)


def supported_provider_names() -> tuple[str, ...]:
    return tuple(provider.name for provider in load_provider_catalog().providers)


def provider_options() -> tuple[tuple[str, str], ...]:
    return tuple(
        (provider.label, provider.name) for provider in load_provider_catalog().providers
    )


def get_provider_catalog(provider_name: str) -> ProviderCatalogEntry:
    for provider in load_provider_catalog().providers:
        if provider.name == provider_name:
            return provider
    raise ValueError("Unsupported provider")


def model_options(provider_name: str) -> tuple[tuple[str, str], ...]:
    provider = get_provider_catalog(provider_name)
    return tuple((model.label, model.id) for model in provider.models)


def supported_models(provider_name: str) -> tuple[str, ...]:
    provider = get_provider_catalog(provider_name)
    return tuple(model.id for model in provider.models)


def all_supported_models() -> tuple[str, ...]:
    models: list[str] = []
    for provider in load_provider_catalog().providers:
        models.extend(model.id for model in provider.models)
    return tuple(models)


def default_model(provider_name: str) -> str:
    return get_provider_catalog(provider_name).default_model


def stale_default_models(provider_name: str) -> tuple[str, ...]:
    return get_provider_catalog(provider_name).stale_default_models


def api_key_env(provider_name: str) -> str:
    return get_provider_catalog(provider_name).api_key_env


def price_for_model(provider_name: str, model: str) -> ModelTokenPrice | None:
    provider = get_provider_catalog(provider_name)
    for entry in provider.models:
        if entry.pricing is not None and model == entry.id:
            return entry.pricing
    for entry in provider.models:
        if entry.pricing is not None and model.startswith(f"{entry.id}-"):
            return entry.pricing
    return None


def price_for_model_any_provider(model: str) -> ModelTokenPrice | None:
    for provider in load_provider_catalog().providers:
        price = price_for_model(provider.name, model)
        if price is not None:
            return price
    return None


def pricing_source() -> str:
    return load_provider_catalog().pricing_source


def _require_str(value: object, message: str) -> str:
    if type(value) is not str:
        raise TypeError(message)
    return value
